from sklearn.datasets import load_breast_cancer, load_digits, load_iris, load_wine
from sklearn.naive_bayes import GaussianNB

from tests.benchmarks.utils import benchmark
from wnb import Distribution as D
from wnb import GaussianWNB, GeneralNB

MAX_ITER = 100


def benchmark_wine() -> None:
    X, y = load_wine(return_X_y=True)

    clf_wnb = GeneralNB(distributions=[D.LOGNORMAL] * X.shape[1])
    clf_sklearn = GaussianNB()
    score_wnb, score_sklearn = benchmark(X, y, clf_wnb, clf_sklearn, MAX_ITER)

    print("wine dataset | sklearn | GaussianNB >> score >>", score_sklearn)
    print("wine dataset | wnb     | GeneralNB  >> score >>", score_wnb, "\n")


def benchmark_iris() -> None:
    X, y = load_iris(return_X_y=True)

    clf_wnb = GeneralNB(distributions=[D.EXPONENTIAL, D.RAYLEIGH, D.NORMAL, D.NORMAL])
    clf_sklearn = GaussianNB()
    score_wnb, score_sklearn = benchmark(X, y, clf_wnb, clf_sklearn, MAX_ITER)

    print("iris dataset | sklearn | GaussianNB >> score >>", score_sklearn)
    print("iris dataset | wnb     | GeneralNB  >> score >>", score_wnb, "\n")


def benchmark_digits() -> None:
    X, y = load_digits(return_X_y=True)

    clf_wnb = GeneralNB(distributions=[D.POISSON] * X.shape[1])
    clf_sklearn = GaussianNB()

    score_wnb, score_sklearn = benchmark(X, y, clf_wnb, clf_sklearn, MAX_ITER)

    print("digits dataset | sklearn | GaussianNB >> score >>", score_sklearn)
    print("digits dataset | wnb     | GeneralNB  >> score >>", score_wnb, "\n")


def benchmark_breast_cancer() -> None:
    X, y = load_breast_cancer(return_X_y=True)

    clf_wnb = GaussianWNB(max_iter=20, step_size=0.01, C=1.5)
    clf_sklearn = GaussianNB()
    score_wnb, score_sklearn = benchmark(X, y, clf_wnb, clf_sklearn, MAX_ITER)

    print("breast cancer dataset | sklearn | GaussianNB  >> score >>", score_sklearn)
    print("breast cancer dataset | wnb     | GaussianWNB >> score >>", score_wnb, "\n")


if __name__ == "__main__":
    benchmark_wine()
    benchmark_iris()
    benchmark_digits()
    benchmark_breast_cancer()



import numpy as np
import pytest


@pytest.fixture
def global_random_seed():
    return np.random.randint(0, 1000)


@pytest.fixture
def random_uniform():
    return np.random.uniform(0, 100, size=10000)



from .base import DistMixin
from .continuous import (
    BetaDist,
    ChiSquaredDist,
    ExponentialDist,
    GammaDist,
    LognormalDist,
    NormalDist,
    ParetoDist,
    RayleighDist,
    TDist,
    UniformDist,
)
from .discrete import BernoulliDist, CategoricalDist, GeometricDist, PoissonDist
from .enums import Distribution

_all_distribution_classes = {
    name: obj
    for name, obj in locals().items()
    if isinstance(obj, type) and issubclass(obj, DistMixin) and obj != DistMixin
}

AllDistributions = {cls.name: cls for cls in _all_distribution_classes.values()}

NonNumericDistributions = [Distribution.CATEGORICAL]

__all__ = [
    "NormalDist",
    "LognormalDist",
    "ExponentialDist",
    "UniformDist",
    "ParetoDist",
    "GammaDist",
    "BetaDist",
    "ChiSquaredDist",
    "TDist",
    "RayleighDist",
    "BernoulliDist",
    "CategoricalDist",
    "GeometricDist",
    "PoissonDist",
    "Distribution",
    "AllDistributions",
    "NonNumericDistributions",
]



from __future__ import annotations

import inspect
import sys
import warnings
from abc import ABCMeta, abstractmethod
from functools import wraps
from numbers import Number
from typing import Any

import numpy as np

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from .enums import Distribution

__all__ = ["ContinuousDistMixin", "DiscreteDistMixin"]


class DistMixin(metaclass=ABCMeta):
    """
    Mixin class for probability distributions in wnb.
    """

    name: str | Distribution
    _support: list[float] | tuple[float, float]

    @classmethod
    @abstractmethod
    def from_data(cls, data, **kwargs: Any) -> Self:
        """Creates an instance of the class from given data. Distribution parameters will be estimated from data.

        Args:
            data: Input data from which distribution parameters will be estimated.

        Returns:
            self: An instance of the class.
        """
        pass

    @classmethod
    def _get_param_names(cls) -> list[str]:
        """
        Gets parameter names for the distribution instance.
        """
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            return []

        init_signature = inspect.signature(init)
        parameters = [
            p for p in init_signature.parameters.values() if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]

        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "wnb estimators should always "
                    "specify their parameters in the signature "
                    "of their __init__ (no varargs). "
                    "%s with constructor %s doesn't "
                    "follow this convention." % (cls, init_signature)
                )

        return sorted([p.name for p in parameters])

    def get_params(self) -> dict[str, Any]:
        """Gets parameters for this distribution instance.

        Returns:
            dict: Parameter names mapped to their values.
        """
        out: dict[str, Any] = {}
        for key in self._get_param_names():
            value = getattr(self, key)
            out[key] = value
        return out

    @property
    def support(self) -> list[float] | tuple[float, float]:
        """Returns the support of the probability distribution.

        If support is a list, it represents a limited number of discrete values.
        If it is a tuple, it indicates a limited or unlimited range of continuous values.
        """
        return self._support

    def _check_support(self, x) -> None:
        if (isinstance(self.support, list) and x not in self.support) or (
            isinstance(self.support, tuple) and (x < self.support[0] or x > self.support[1])
        ):
            warnings.warn(
                "Value doesn't lie within the support of the distribution",
                RuntimeWarning,
            )

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass

    def __repr__(self) -> str:
        return "".join(
            [
                "<",
                self.__class__.__name__,
                "(",
                ", ".join(
                    [
                        f"{k}={v:.4f}" if isinstance(v, Number) else f"{k}={v}"
                        for k, v in self.get_params().items()
                    ]
                ),
                ")>",
            ]
        )


def vectorize(
    otypes: str | None = None,
    excluded: set[int | str] = None,
    signature: str | None = None,
):
    """
    Numpy vectorization wrapper that works with class methods.
    """

    def decorator(func):
        vectorized = np.vectorize(func, otypes=otypes, excluded=excluded, signature=signature)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any):
            return vectorized(*args, **kwargs)

        return wrapper

    return decorator


class ContinuousDistMixin(DistMixin, metaclass=ABCMeta):
    """
    Mixin class for all continuous probability distributions in wnb.
    """

    _type: str = "continuous"

    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes an instance of the continuous probability distribution with given parameters.
        """
        pass

    @abstractmethod
    def pdf(self, x: float) -> float:
        """Returns the value of probability density function (PDF) at x.

        Args:
            x (float): Input value.

        Returns:
            float: Probability density.
        """
        pass

    @vectorize(signature="(),()->()")
    def __call__(self, x: float) -> float:
        self._check_support(x)
        return self.pdf(x)


class DiscreteDistMixin(DistMixin, metaclass=ABCMeta):
    """
    Mixin class for all discrete probability distributions in wnb.
    """

    _type: str = "discrete"

    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes an instance of the discrete probability distribution with given parameters.
        """
        pass

    @abstractmethod
    def pmf(self, x: float) -> float:
        """Returns the value of probability mass function (PMF) at x.

        Args:
            x (float): Input value.

        Returns:
            float: Probability mass.
        """
        pass

    @vectorize(signature="(),()->()")
    def __call__(self, x: float) -> float:
        self._check_support(x)
        return self.pmf(x)



import numpy as np
import pytest
from scipy import stats
from sklearn.utils._testing import assert_array_almost_equal

from wnb import Distribution as D
from wnb.stats import (
    AllDistributions,
    BernoulliDist,
    BetaDist,
    ChiSquaredDist,
    ExponentialDist,
    GammaDist,
    GeometricDist,
    LognormalDist,
    NormalDist,
    ParetoDist,
    PoissonDist,
    RayleighDist,
    TDist,
    UniformDist,
)

out_of_support_warn_msg = "Value doesn't lie within the support of the distribution"


def test_distributions_correct_name_attr():
    """
    Test if all defined distributions have correct `name` attributes.
    """
    for dist_name in AllDistributions.keys():
        assert isinstance(dist_name, (str, D))


def test_distributions_correct_support_attr():
    """
    Test if all defined distributions have correct `_support` attributes.
    """
    for dist in AllDistributions.values():
        if dist.name in [D.UNIFORM, D.PARETO, D.CATEGORICAL]:
            assert dist._support is None
            continue

        assert isinstance(dist._support, (list, tuple))

        if isinstance(dist._support, list):
            for x in dist._support:
                assert isinstance(x, (float, int))
        else:
            assert len(dist._support) == 2
            for x in dist._support:
                assert isinstance(x, (float, int))


def test_normal_pdf():
    """
    Test whether pdf method of `NormalDist` returns the same result as pdf method of `scipy.stats.norm`.
    """
    norm_wnb = NormalDist(mu=1, sigma=3)
    norm_scipy = stats.norm(loc=1, scale=3)
    X = np.random.uniform(-100, 100, size=10000)
    assert_array_almost_equal(norm_wnb(X), norm_scipy.pdf(X), decimal=10)


@pytest.mark.parametrize("epsilon", [1e-10, 1e-9, 1e-6, 1e-3])
def test_normal_with_epsilon(epsilon: float):
    """
    Test whether epsilon is correctly applied for `NormalDist`.
    """
    norm_1 = NormalDist(mu=1, sigma=0)
    norm_2 = NormalDist(mu=1, sigma=0, epsilon=epsilon)
    norm_3 = NormalDist(mu=1, sigma=np.sqrt(epsilon))
    assert norm_1.sigma == norm_2.sigma == 0
    assert norm_3.sigma == np.sqrt(epsilon)
    X = np.random.uniform(-100, 100, size=10000)
    assert np.isnan(norm_1(X)).all()
    assert_array_almost_equal(norm_2(X), norm_3(X), decimal=10)


def test_lognormal_pdf(random_uniform):
    """
    Test whether pdf method of `LognormalDist` returns the same result as pdf method of `scipy.stats.lognorm`.
    """
    lognorm_wnb = LognormalDist(mu=1, sigma=3)
    lognorm_scipy = stats.lognorm(scale=np.exp(1), s=3)
    X = random_uniform
    assert_array_almost_equal(lognorm_wnb(X), lognorm_scipy.pdf(X), decimal=10)


def test_lognormal_out_of_support_data():
    """
    Test whether a warning is issued when calling `LognormalDist` with out-of-support data.
    """
    X = np.random.lognormal(mean=1.5, sigma=3.75, size=1000)
    lognorm_wnb = LognormalDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        lognorm_wnb(-1)


def test_exponential_pdf(random_uniform):
    """
    Test whether pdf method of `ExponentialDist` returns the same result as pdf method of `scipy.stats.expon`.
    """
    expon_wnb = ExponentialDist(rate=4)
    expon_scipy = stats.expon(scale=1 / 4)
    X = random_uniform
    assert_array_almost_equal(expon_wnb(X), expon_scipy.pdf(X), decimal=10)


def test_exponential_out_of_support_data():
    """
    Test whether a warning is issued when calling `ExponentialDist` with out-of-support data.
    """
    X = np.random.exponential(scale=2.5, size=1000)
    expon_wnb = ExponentialDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        expon_wnb(-1)


def test_uniform_pdf(random_uniform):
    """
    Test whether pdf method of `UniformDist` returns the same result as pdf method of `scipy.stats.uniform`.
    """
    uniform_wnb = UniformDist(a=-5, b=10)
    uniform_scipy = stats.uniform(loc=-5, scale=15)
    X = random_uniform
    assert_array_almost_equal(uniform_wnb(X), uniform_scipy.pdf(X), decimal=10)


def test_uniform_out_of_support_data():
    """
    Test whether a warning is issued when calling `UniformDist` with out-of-support data.
    """
    X = np.random.uniform(low=5, high=10, size=1000)
    uniform_wnb = UniformDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        uniform_wnb(3)


def test_pareto_pdf(random_uniform):
    """
    Test whether pdf method of `ParetoDist` returns the same result as pdf method of `scipy.stats.pareto`.
    """
    pareto_wnb = ParetoDist(x_m=5, alpha=0.5)
    pareto_scipy = stats.pareto(b=0.5, scale=5)
    X = random_uniform
    assert_array_almost_equal(pareto_wnb(X), pareto_scipy.pdf(X), decimal=10)


def test_pareto_out_of_support_data():
    """
    Test whether a warning is issued when calling `ParetoDist` with out-of-support data.
    """
    X = np.random.pareto(a=0.5, size=1000)
    pareto_wnb = ParetoDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        pareto_wnb(-5)


def test_gamma_pdf(random_uniform):
    """
    Test whether pdf method of `GammaDist` returns the same result as pdf method of `scipy.stats.gamma`.
    """
    gamma_wnb = GammaDist(k=1, theta=3)
    gamma_scipy = stats.gamma(a=1, scale=3)
    X = random_uniform
    assert_array_almost_equal(gamma_wnb(X), gamma_scipy.pdf(X), decimal=10)


def test_gamma_out_of_support_data():
    """
    Test whether a warning is issued when calling `GammaDist` with out-of-support data.
    """
    X = np.random.gamma(shape=1, scale=3, size=1000)
    gamma_wnb = GammaDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        gamma_wnb(-5)


def test_beta_pdf():
    """
    Test whether pdf method of `BetaDist` returns the same result as pdf method of `scipy.stats.beta`.
    """
    beta_wnb = BetaDist(alpha=1, beta=5)
    beta_scipy = stats.beta(a=1, b=5)
    X = np.random.uniform(0.01, 0.99, size=10000)
    assert_array_almost_equal(beta_wnb(X), beta_scipy.pdf(X), decimal=10)


def test_beta_out_of_support_data():
    """
    Test whether a warning is issued when calling `BetaDist` with out-of-support data.
    """
    X = np.random.beta(a=1, b=5, size=1000)
    beta_wnb = BetaDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        beta_wnb(1.01)


def test_chi2_pdf(random_uniform):
    """
    Test whether pdf method of `ChiSquaredDist` returns the same result as pdf method of `scipy.stats.chi2`.
    """
    chi2_wnb = ChiSquaredDist(k=4)
    chi2_scipy = stats.chi2(df=4)
    X = random_uniform
    assert_array_almost_equal(chi2_wnb(X), chi2_scipy.pdf(X), decimal=10)


def test_chi2_out_of_support_data():
    """
    Test whether a warning is issued when calling `ChiSquaredDist` with out-of-support data.
    """
    X = np.random.chisquare(df=4, size=1000)
    chi2_wnb = ChiSquaredDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        chi2_wnb(-5)


def test_t_pdf(random_uniform):
    """
    Test whether pdf method of `TDist` returns the same result as pdf method of `scipy.stats.t`.
    """
    t_wnb = TDist(df=10)
    t_scipy = stats.t(df=10)
    X = random_uniform
    assert_array_almost_equal(t_wnb(X), t_scipy.pdf(X), decimal=10)


def test_rayleigh_pdf(random_uniform):
    """
    Test whether pdf method of `RayleighDist` returns the same result as pdf method of `scipy.stats.rayleigh`.
    """
    rayleigh_wnb = RayleighDist(sigma=5)
    rayleigh_scipy = stats.rayleigh(scale=5)
    X = random_uniform
    assert_array_almost_equal(rayleigh_wnb(X), rayleigh_scipy.pdf(X), decimal=10)


def test_rayleigh_out_of_support_data():
    """
    Test whether a warning is issued when calling `RayleighDist` with out-of-support data.
    """
    X = np.random.rayleigh(scale=4, size=1000)
    rayleigh_wnb = RayleighDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        rayleigh_wnb(-5)


def test_bernoulli_pdf():
    """
    Test whether pmf method of `BernoulliDist` returns the same result as pmf method of `scipy.stats.bernoulli`.
    """
    bernoulli_wnb = BernoulliDist(p=0.4)
    bernoulli_scipy = stats.bernoulli(p=0.4)
    X = np.random.randint(0, 2, size=10000)
    assert_array_almost_equal(bernoulli_wnb(X), bernoulli_scipy.pmf(X), decimal=10)


def test_bernoulli_out_of_support_data():
    """
    Test whether a warning is issued when calling `BernoulliDist` with out-of-support data.
    """
    X = np.random.randint(0, 2, size=1000)
    bernoulli_wnb = BernoulliDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        bernoulli_wnb(2)


def test_geometric_pdf():
    """
    Test whether pmf method of `GeometricDist` returns the same result as pmf method of `scipy.stats.geom`.
    """
    geom_wnb = GeometricDist(p=0.3)
    geom_scipy = stats.geom(p=0.3)
    X = np.random.randint(1, 1000, size=10000)
    assert_array_almost_equal(geom_wnb(X), geom_scipy.pmf(X), decimal=10)


def test_geometric_out_of_support_data():
    """
    Test whether a warning is issued when calling `GeometricDist` with out-of-support data.
    """
    X = np.random.geometric(p=0.3, size=1000)
    geom_wnb = GeometricDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        geom_wnb(0)


def test_poisson_pdf():
    """
    Test whether pmf method of `PoissonDist` returns the same result as pmf method of `scipy.stats.poisson`.
    """
    poisson_wnb = PoissonDist(rate=1.5)
    poisson_scipy = stats.poisson(mu=1.5)
    X = np.random.randint(0, 100, size=10000)
    assert_array_almost_equal(poisson_wnb(X), poisson_scipy.pmf(X), decimal=10)


def test_poisson_out_of_support_data():
    """
    Test whether a warning is issued when calling `PoissonDist` with out-of-support data.
    """
    X = np.random.poisson(lam=1.5, size=1000)
    poisson_wnb = PoissonDist.from_data(X)

    with pytest.warns(RuntimeWarning, match=out_of_support_warn_msg):
        poisson_wnb(-1)



from typing import Type, Union

from .base import ContinuousDistMixin, DiscreteDistMixin
from .enums import Distribution

__all__ = ["DistributionLike"]


DistributionLike = Union[
    str,
    Distribution,
    Type[ContinuousDistMixin],
    Type[DiscreteDistMixin],
]



import warnings
from typing import Union

import numpy as np
import numpy.typing
import pandas as pd
from scipy.sparse import spmatrix

__all__ = ["MatrixLike", "ArrayLike", "Int", "Float", "DistributionLike"]  # noqa: F822


ArrayLike = numpy.typing.ArrayLike
MatrixLike = Union[np.ndarray, pd.DataFrame, spmatrix]

Int = Union[int, np.int8, np.int16, np.int32, np.int64]
Float = Union[float, np.float16, np.float32, np.float64]


def __getattr__(name):
    if name == "DistributionLike":
        warnings.warn(
            "The `DistributionLike` type has been moved to `wnb.stats.typing`. "
            "Please update your imports accordingly. This import will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )
        from wnb.stats.typing import DistributionLike

        return DistributionLike
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



import warnings

from wnb.stats import *  # noqa: F403

warnings.warn(
    "The `wnb.dist` module is deprecated and will be removed in a future release. "
    "Please update your imports to use `wnb` or `wnb.stats` directly. "
    "Using `wnb.dist` will continue to work in this version, but it may be removed in future versions.",
    DeprecationWarning,
    stacklevel=2,
)



from sklearn.datasets import load_wine
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB

from wnb import Distribution as D
from wnb import GeneralNB

# Load the wine dataset
wine = load_wine()
X = wine["data"]
y = wine["target"]

# Split the data into training and test sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=0)

# Train and score sklearn GaussianNB classifier
gnb = GaussianNB()
gnb.fit(X_train, y_train)
print("sklearn | GaussianNB >> score >>", gnb.score(X_test, y_test))

# Train and score wnb GeneralNB classifier with Log-normal likelihoods
gnb = GeneralNB(distributions=[D.LOGNORMAL] * X.shape[1])
gnb.fit(X_train, y_train)
print("wnb | GeneralNB >> score >>", gnb.score(X_test, y_test))



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB

from wnb import GaussianWNB

# Load the breast cancer wisconsin dataset
breast_cancer = load_breast_cancer()
X = breast_cancer["data"]
y = breast_cancer["target"]

# Split the data into training and test sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=0)

# Train and score sklearn GaussianNB classifier
gnb = GaussianNB()
gnb.fit(X_train, y_train)
print("sklearn | GaussianNB >> score >>", gnb.score(X_test, y_test))

# Train and score wnb GaussianWNB classifier
gwnb = GaussianWNB(max_iter=20, step_size=0.01)
gwnb.fit(X_train, y_train)
print("wnb | GaussianWNB >> score >>", gwnb.score(X_test, y_test))



from math import factorial
from typing import Any, Mapping

import numpy as np

from .base import DiscreteDistMixin
from .enums import Distribution as D

__all__ = [
    "BernoulliDist",
    "CategoricalDist",
    "GeometricDist",
    "PoissonDist",
]


class BernoulliDist(DiscreteDistMixin):
    name = D.BERNOULLI
    _support = [0, 1]

    def __init__(self, p: float) -> None:
        self.p = p
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "BernoulliDist":
        alpha = kwargs.get("alpha", 1e-10)
        return cls(p=((np.array(data) == 1).sum() + alpha) / len(data))

    def pmf(self, x: int) -> float:
        if x not in self._support:
            return 0.0
        else:
            return self.p if x == 1 else 1 - self.p


class CategoricalDist(DiscreteDistMixin):
    name = D.CATEGORICAL
    _support = None

    def __init__(self, prob: Mapping[Any, float]) -> None:
        self.prob = prob
        self._support = list(self.prob.keys())
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "CategoricalDist":
        alpha = kwargs.get("alpha", 1e-10)
        values, counts = np.unique(data, return_counts=True)
        return cls(prob={v: (c + alpha) / len(data) for v, c in zip(values, counts)})

    def pmf(self, x: Any) -> float:
        return self.prob.get(x, 0.0)


class GeometricDist(DiscreteDistMixin):
    name = D.GEOMETRIC
    _support = (1, np.inf)

    def __init__(self, p: float) -> None:
        self.p = p
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "GeometricDist":
        return cls(p=len(data) / np.sum(data))

    def pmf(self, x: int) -> float:
        return self.p * (1 - self.p) ** (x - 1) if x >= self._support[0] and x - int(x) == 0 else 0.0


class PoissonDist(DiscreteDistMixin):
    name = D.POISSON
    _support = (0, np.inf)

    def __init__(self, rate: float) -> None:
        self.rate = rate
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "PoissonDist":
        return cls(rate=np.sum(data) / len(data))

    def pmf(self, x: int) -> float:
        return (
            (np.exp(-self.rate) * self.rate**x) / factorial(x)
            if x >= self._support[0] and x - int(x) == 0
            else 0.0
        )



from typing import Any, Optional

import numpy as np
from scipy.special import beta, gamma
from scipy.stats import chi2

from .base import ContinuousDistMixin
from .enums import Distribution as D

__all__ = [
    "NormalDist",
    "LognormalDist",
    "ExponentialDist",
    "UniformDist",
    "ParetoDist",
    "GammaDist",
    "BetaDist",
    "ChiSquaredDist",
    "TDist",
    "RayleighDist",
]


class NormalDist(ContinuousDistMixin):
    name = D.NORMAL
    _support = (-np.inf, np.inf)

    def __init__(self, mu: float, sigma: float, epsilon: Optional[float] = None) -> None:
        self.mu = mu
        self.sigma = sigma
        self.epsilon = epsilon
        super().__init__()

    @classmethod
    def from_data(cls, data: np.ndarray, **kwargs: Any) -> "NormalDist":
        epsilon = kwargs.get("epsilon")
        return cls(mu=np.average(data), sigma=np.std(data), epsilon=epsilon)

    def pdf(self, x: float) -> float:
        var = self.sigma**2 + (self.epsilon or 0.0)
        return (1.0 / np.sqrt(2 * np.pi * var)) * np.exp(-0.5 * ((x - self.mu) ** 2 / var))


class LognormalDist(ContinuousDistMixin):
    name = D.LOGNORMAL
    _support = (0, np.inf)

    def __init__(self, mu: float, sigma: float) -> None:
        self.mu = mu
        self.sigma = sigma
        super().__init__()

    @classmethod
    def from_data(cls, data: np.ndarray, **kwargs: Any) -> "LognormalDist":
        log_data = np.log(data)
        return cls(mu=np.average(log_data), sigma=np.std(log_data))

    def pdf(self, x: float) -> float:
        return (
            (1.0 / (x * self.sigma * np.sqrt(2 * np.pi)))
            * np.exp(-0.5 * ((np.log(x) - self.mu) / self.sigma) ** 2)
            if x > 0
            else 0.0
        )


class ExponentialDist(ContinuousDistMixin):
    name = D.EXPONENTIAL
    _support = (0, np.inf)

    def __init__(self, rate: float) -> None:
        self.rate = rate
        super().__init__()

    @classmethod
    def from_data(cls, data: np.ndarray, **kwargs: Any) -> "ExponentialDist":
        return cls(rate=(len(data) - 2) / np.sum(data))

    def pdf(self, x: float) -> float:
        return self.rate * np.exp(-self.rate * x) if x >= 0 else 0.0


class UniformDist(ContinuousDistMixin):
    name = D.UNIFORM
    _support = None

    def __init__(self, a: float, b: float) -> None:
        self.a = a
        self.b = b
        self._support = (a, b)
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "UniformDist":
        return cls(a=np.min(data), b=np.max(data))

    def pdf(self, x: float) -> float:
        return 1 / (self.b - self.a) if self.a <= x <= self.b else 0.0


class ParetoDist(ContinuousDistMixin):
    name = D.PARETO
    _support = None

    def __init__(self, x_m: float, alpha: float) -> None:
        self.x_m = x_m
        self.alpha = alpha
        self._support = (self.x_m, np.inf)
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "ParetoDist":
        x_m = np.min(data)
        return cls(x_m=x_m, alpha=len(data) / np.sum(np.log(data / x_m)))

    def pdf(self, x: float) -> float:
        return (self.alpha * self.x_m**self.alpha) / x ** (self.alpha + 1) if x >= self.x_m else 0.0


class GammaDist(ContinuousDistMixin):
    name = D.GAMMA
    _support = (0, np.inf)

    def __init__(self, k: float, theta: float) -> None:
        self.k = k
        self.theta = theta
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "GammaDist":
        n = len(data)
        return cls(
            k=n * np.sum(data) / (n * np.sum(data * np.log(data)) - np.sum(data * np.sum(np.log(data)))),
            theta=(n * np.sum(data * np.log(data)) - np.sum(data * np.sum(np.log(data)))) / n**2,
        )

    def pdf(self, x: float) -> float:
        return (x ** (self.k - 1) * np.exp(-x / self.theta)) / (gamma(self.k) * self.theta**self.k)


class BetaDist(ContinuousDistMixin):
    name = D.BETA
    _support = (0, 1)

    def __init__(self, alpha: float, beta: float) -> None:
        self.alpha = alpha
        self.beta = beta
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "BetaDist":
        mu_hat = np.average(data)
        var_hat = np.var(data, ddof=1)
        multiplied_term = (mu_hat * (1 - mu_hat) / var_hat) - 1
        return cls(
            alpha=mu_hat * multiplied_term,
            beta=(1 - mu_hat) * multiplied_term,
        )

    def pdf(self, x: float) -> float:
        return ((x ** (self.alpha - 1)) * (1 - x) ** (self.beta - 1)) / beta(self.alpha, self.beta)


class ChiSquaredDist(ContinuousDistMixin):
    name = D.CHI_SQUARED
    _support = (0, np.inf)

    def __init__(self, k: int) -> None:
        self.k = k
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "ChiSquaredDist":
        return cls(k=round(np.average(data)))

    def pdf(self, x: float) -> float:
        return chi2.pdf(x, self.k)


class TDist(ContinuousDistMixin):
    name = D.T
    _support = (-np.inf, np.inf)

    def __init__(self, df: float) -> None:
        self.df = df
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "TDist":
        return cls(df=len(data) - 1)

    def pdf(self, x: float) -> float:
        return (gamma((self.df + 1) / 2) / (np.sqrt(self.df * np.pi) * gamma(self.df / 2))) * (
            1 + (x**2 / self.df)
        ) ** (-(self.df + 1) / 2)


class RayleighDist(ContinuousDistMixin):
    name = D.RAYLEIGH
    _support = (0, np.inf)

    def __init__(self, sigma: float) -> None:
        self.sigma = sigma
        super().__init__()

    @classmethod
    def from_data(cls, data, **kwargs: Any) -> "RayleighDist":
        sigma = np.sqrt(np.mean(data**2) / 2)
        return cls(sigma=sigma)

    def pdf(self, x: float) -> float:
        return (x / self.sigma**2) * np.exp(-(x**2) / (2 * self.sigma**2)) if x >= 0 else 0.0



"""
Python library for the implementations of general and weighted naive Bayes (WNB) classifiers.
"""

__version__ = "0.5.1"
__author__ = "Mehdi Samsami"


from wnb.gnb import GeneralNB
from wnb.gwnb import GaussianWNB
from wnb.stats.base import ContinuousDistMixin, DiscreteDistMixin
from wnb.stats.enums import Distribution

__all__ = [
    "GeneralNB",
    "GaussianWNB",
    "Distribution",
    "ContinuousDistMixin",
    "DiscreteDistMixin",
]



from __future__ import annotations

import numbers
import sys
import warnings
from abc import ABCMeta
from typing import Optional

import numpy as np
import pandas as pd
from scipy.special import logsumexp
from scipy.stats import norm
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils import as_float_array
from sklearn.utils.multiclass import check_classification_targets, type_of_target
from sklearn.utils.validation import _ensure_no_complex_data, check_is_fitted

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ._utils import (
    SKLEARN_V1_6_OR_LATER,
    _check_feature_names,
    _check_n_features,
    validate_data,
)
from .typing import ArrayLike, Float, Int, MatrixLike

__all__ = ["GaussianWNB"]


class GaussianWNB(ClassifierMixin, BaseEstimator, metaclass=ABCMeta):
    """Binary Gaussian Minimum Log-likelihood Difference Weighted Naive Bayes (MLD-WNB) classifier

    Parameters
    ----------
    priors : array-like of shape (n_classes,), default=None
        Prior probabilities of the classes. If specified, the priors are not
        adjusted according to the data.

    error_weights : array-like of shape (n_classes, n_classes), default=None
        Matrix of error weights. If not specified, equal weight is assigned to the
        errors of both classes.

    max_iter : int, default=25
        Maximum number of gradient descent iterations.

    step_size : float, default=1e-4
        Step size of weight update (i.e., learning rate).

    penalty : str, default="l2"
        Regularization term, either 'l1' or 'l2'.

    C : float, default=1.0
        Regularization strength. Must be strictly positive.

    learning_hist : bool, default=False
        Whether to record the learning history, i.e., the value of cost function
        in each learning iteration.

    Attributes
    ----------
    class_count_ : ndarray of shape (n_classes,)
        Number of training samples observed in each class.

    class_prior_ : ndarray of shape (n_classes,)
        Probability of each class.

    classes_ : ndarray of shape (n_classes,)
        Class labels known to the classifier.

    n_classes_ : int
        Number of classes seen during :term:`fit`.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    error_weights_ : ndarray of shape (n_classes, n_classes)
        Specified matrix of error weights.

    theta_ : ndarray of shape (n_features, n_classes)
        Mean of each feature per class.

    std_ : ndarray of shape (n_features, n_classes)
        Standard deviation of each feature per class.

    var_ : ndarray of shape (n_features, n_classes)
        Variance of each feature per class.

    coef_ : ndarray of shape (n_features,)
        Weights assigned to the features.

    cost_hist_ : ndarray of shape (max_iter,)
        Cost value in each iteration of the optimization.

    n_iter_ : int
        Number of iterations run by the optimization routine to fit the model.
    """

    def __init__(
        self,
        *,
        priors: Optional[ArrayLike] = None,
        error_weights: Optional[np.ndarray] = None,
        max_iter: Int = 25,
        step_size: Float = 1e-4,
        penalty: str = "l2",
        C: Float = 1.0,
        learning_hist: bool = False,
    ) -> None:
        self.priors = priors
        self.error_weights = error_weights
        self.max_iter = max_iter
        self.step_size = step_size
        self.penalty = penalty
        self.C = C
        self.learning_hist = learning_hist

    if SKLEARN_V1_6_OR_LATER:

        def __sklearn_tags__(self):
            tags = super().__sklearn_tags__()
            tags.target_tags.required = True
            tags.classifier_tags.multi_class = False
            return tags

    def _more_tags(self) -> dict[str, bool]:
        return {"binary_only": True, "requires_y": True}

    def _check_inputs(self, X, y) -> None:
        # Check if the targets are suitable for classification
        check_classification_targets(y)

        # Check that the dataset has only two unique labels
        if (y_type := type_of_target(y)) != "binary":
            if SKLEARN_V1_6_OR_LATER:
                msg = f"Only binary classification is supported. The type of the target is {y_type}."
            else:
                msg = "Unknown label type: non-binary"
            raise ValueError(msg)

        # Check if only one class is present in label vector
        if self.n_classes_ == 1:
            raise ValueError("Classifier can't train when only one class is present.")

        X = validate_data(
            self,
            X,
            accept_sparse=False,
            accept_large_sparse=False,
            dtype="numeric",
            force_all_finite=True,
            ensure_2d=True,
            ensure_min_samples=1,
            ensure_min_features=1,
        )

        # Check that the number of samples and labels are compatible
        if self.__n_samples != y.shape[0]:
            raise ValueError("X.shape[0]=%d and y.shape[0]=%d are incompatible." % (X.shape[0], y.shape[0]))

        if self.priors is not None:
            # Check that the provided priors match the number of classes
            if len(self.priors) != self.n_classes_:
                raise ValueError("Number of priors must match the number of classes.")
            # Check that the sum of priors is 1
            if not np.isclose(self.priors.sum(), 1.0):
                raise ValueError("The sum of the priors should be 1.")
            # Check that the priors are non-negative
            if (self.priors < 0).any():
                raise ValueError("Priors must be non-negative.")

        if self.error_weights is not None:
            # Check that the size of error weights matrix matches number of classes
            if self.error_weights.shape != (self.n_classes_, self.n_classes_):
                raise ValueError(
                    "The shape of error weights matrix does not match the number of classes, "
                    "must be (n_classes, n_classes)."
                )

        # Check that the regularization type is either 'l1' or 'l2'
        if self.penalty not in ["l1", "l2"]:
            raise ValueError("Regularization type must be either 'l1' or 'l2'.")

        # Check that the regularization parameter is a positive integer
        if not isinstance(self.C, numbers.Number) or self.C < 0:
            raise ValueError("Regularization parameter must be positive; got (C=%r)" % self.C)

        # Check that the maximum number of iterations is a positive integer
        if not isinstance(self.max_iter, numbers.Number) or self.max_iter < 0:
            raise ValueError(
                "Maximum number of iteration must be a positive integer; got (max_iter=%r)." % self.max_iter
            )

    def _prepare_X_y(self, X=None, y=None, from_fit: bool = False):
        if from_fit and y is None:
            raise ValueError("requires y to be passed, but the target y is None.")

        if X is not None:
            # Convert to NumPy array if X is Pandas DataFrame
            if isinstance(X, pd.DataFrame):
                X = X.values
            _ensure_no_complex_data(X)
            X = as_float_array(X)

        if y is not None:
            # Convert to a NumPy array
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
            else:
                y = np.array(y)

            # Warning in case of y being 2d
            if y.ndim > 1:
                warnings.warn(
                    "A column-vector y was passed when a 1d array was expected.",
                    DataConversionWarning,
                )

            y = y.flatten()

        output = tuple(item for item in [X, y] if item is not None)
        return output[0] if len(output) == 1 else output

    def _prepare_parameters(self, X, y) -> None:
        # Calculate mean and standard deviation of features for each class
        for c in range(self.n_classes_):
            self.theta_[:, c] = np.mean(X[y == c, :], axis=0)  # Calculate mean of features for class c
            self.std_[:, c] = np.std(X[y == c, :], axis=0)  # Calculate std of features for class c
        self.var_ = np.square(self.std_)  # Calculate variance of features using std

        self.class_prior_: np.ndarray
        # Update if no priors is provided
        if self.priors is None:
            self.class_prior_ = (
                self.class_count_ / self.class_count_.sum()
            )  # Calculate empirical prior probabilities
        else:
            self.class_prior_ = self.priors

        # Convert to NumPy array if input priors is in a list/tuple/set
        if isinstance(self.class_prior_, (list, tuple, set)):
            self.class_prior_ = np.array(list(self.class_prior_))

        # Update if no error weights is provided
        if self.error_weights is None:
            self.error_weights_ = np.array([[0, 1], [-1, 0]])
        else:
            self.error_weights_ = self.error_weights

    def fit(self, X: MatrixLike, y: ArrayLike) -> Self:
        """Fits Gaussian Binary MLD-WNB classifier according to X, y.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vectors, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.n_features_in_: int
        self.feature_names_in_: np.ndarray
        _check_n_features(self, X=X, reset=True)
        _check_feature_names(self, X=X, reset=True)

        X, y = self._prepare_X_y(X, y, from_fit=True)

        self.classes_: np.ndarray
        self.class_count_: np.ndarray
        self.classes_, y_, self.class_count_ = np.unique(
            y, return_counts=True, return_inverse=True
        )  # Unique class labels, their indices, and class counts
        self.n_classes_: int = len(self.classes_)  # Number of classes

        self.__n_samples = X.shape[0]  # Number of samples (for internal use)

        self._check_inputs(X, y)
        y = y_

        self.theta_: np.ndarray = np.zeros(
            (self.n_features_in_, self.n_classes_)
        )  # Mean of each feature per class (n_features x n_classes)
        self.std_: np.ndarray = np.zeros(
            (self.n_features_in_, self.n_classes_)
        )  # Standard deviation of each feature per class (n_features x n_classes)
        self.var_: np.ndarray = np.zeros(
            (self.n_features_in_, self.n_classes_)
        )  # Variance of each feature per class (n_features x n_classes)
        self.coef_: np.ndarray = np.ones((self.n_features_in_,))  # WNB coefficients (n_features x 1)
        self.cost_hist_: np.ndarray = np.array(
            [np.nan for _ in range(self.max_iter)]
        )  # Cost value in each iteration

        self._prepare_parameters(X, y)

        # Learn the weights using gradient descent
        self.n_iter_: int = 0
        for self.n_iter_ in range(self.max_iter):
            # Predict on X
            y_hat = self._predict(X)

            # Calculate cost
            self.cost_hist_[self.n_iter_], _lambda = self._calculate_cost(X, y, y_hat, self.learning_hist)

            # Calculate gradients (most time-consuming)
            _grad = self._calculate_grad(X, _lambda)

            # Add regularization
            if self.penalty == "l1":
                _grad += self.C * np.sign(self.coef_)
            elif self.penalty == "l2":
                _grad += 2 * self.C * self.coef_

            # Update weights
            self.coef_ = self.coef_ - self.step_size * _grad

        self.n_iter_ += 1
        self.cost_hist_ = None if not self.learning_hist else self.cost_hist_

        return self

    def _calculate_cost(self, X, y, y_hat, learning_hist: bool) -> tuple[Float, list[Float]]:
        _lambda = [self.error_weights_[y[i], y_hat[i]] for i in range(self.__n_samples)]

        if learning_hist:
            # Calculate cost
            _cost = 0.0
            for i in range(self.__n_samples):
                _sum = np.log(self.class_prior_[1] / self.class_prior_[0])
                x = X[i, :]
                for j in range(self.n_features_in_):
                    _sum += self.coef_[j] * (
                        np.log(1e-20 + norm.pdf(x[j], self.theta_[j, 1], self.std_[j, 1]))
                        - np.log(1e-20 + norm.pdf(x[j], self.theta_[j, 0], self.std_[j, 0]))
                    )
                _cost += _lambda[i] * _sum
        else:
            _cost = np.nan

        return _cost, _lambda

    def _calculate_grad(self, X, _lambda: list[Float]) -> np.ndarray:
        _grad = np.repeat(
            np.log(self.std_[:, 0] / self.std_[:, 1]).reshape(1, -1),
            self.__n_samples,
            axis=0,
        )
        _grad += (
            0.5
            * (
                (X - np.repeat(self.theta_[:, 0].reshape(1, -1), self.__n_samples, axis=0))
                / (np.repeat(self.std_[:, 0].reshape(1, -1), self.__n_samples, axis=0))
            )
            ** 2
        )
        _grad -= (
            0.5
            * (
                (X - np.repeat(self.theta_[:, 1].reshape(1, -1), self.__n_samples, axis=0))
                / (np.repeat(self.std_[:, 1].reshape(1, -1), self.__n_samples, axis=0))
            )
            ** 2
        )
        _grad *= np.transpose(np.repeat(np.array(_lambda).reshape(1, -1), self.n_features_in_, axis=0))
        _grad = np.sum(_grad, axis=0)

        return _grad

    def _predict(self, X: MatrixLike) -> np.ndarray:
        p_hat = self.predict_log_proba(X)
        return np.argmax(p_hat, axis=1)

    def predict(self, X: MatrixLike) -> np.ndarray:
        """Performs classification on an array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : ndarray of shape (n_samples,)
            Predicted target values for X.
        """
        p_hat = self.predict_log_proba(X)
        y_hat = self.classes_[np.argmax(p_hat, axis=1)]
        return y_hat

    def predict_log_proba(self, X: MatrixLike) -> np.ndarray:
        """Returns log-probability estimates for the array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : array-like of shape (n_samples, n_classes)
            Returns the log-probability of the samples for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute :term:`classes_`.
        """
        # Check is fit had been called
        check_is_fitted(self)

        # Input validation
        X = validate_data(self, X, accept_large_sparse=False, force_all_finite=True, reset=False)

        # Check if the number of input features matches the data seen during fit
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Expected input with %d features, got %d instead." % (self.n_features_in_, X.shape[1])
            )

        n_samples = X.shape[0]

        X = self._prepare_X_y(X=X)

        log_priors = np.tile(np.log(self.class_prior_), (n_samples, 1))
        w_reshaped = np.tile(self.coef_.reshape(-1, 1), (1, self.n_classes_))
        term1 = np.sum(np.multiply(w_reshaped, -np.log(np.sqrt(2 * np.pi) * self.std_)))
        var_inv = np.multiply(w_reshaped, 1.0 / np.multiply(self.std_, self.std_))
        mu_by_var = np.multiply(self.theta_, var_inv)
        term2 = -0.5 * (
            np.matmul(np.multiply(X, X), var_inv)
            - 2.0 * np.matmul(X, mu_by_var)
            + np.sum(self.theta_.conj() * mu_by_var, axis=0)
        )
        log_proba = log_priors + term1 + term2

        log_proba -= np.transpose(
            np.repeat(logsumexp(log_proba, axis=1).reshape(1, -1), self.n_classes_, axis=0)
        )
        return log_proba

    def predict_proba(self, X: MatrixLike) -> np.ndarray:
        """Returns probability estimates for the array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : array-like of shape (n_samples, n_classes)
            Returns the probability of the samples for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute :term:`classes_`.
        """
        return np.exp(self.predict_log_proba(X))



import contextlib
from typing import Optional

from . import AllDistributions
from .base import DistMixin
from .enums import Distribution
from .typing import DistributionLike


def is_dist_supported(dist: DistributionLike) -> bool:
    with contextlib.suppress(TypeError):
        return issubclass(dist, DistMixin)

    if (
        isinstance(dist, Distribution)
        or dist in Distribution.__members__.values()
        or all(hasattr(dist, attr_name) for attr_name in ["from_data", "support", "__call__"])
    ):
        return True

    return False


def get_dist_class(name_or_type: DistributionLike) -> Optional[DistMixin]:
    with contextlib.suppress(TypeError):
        if issubclass(name_or_type, DistMixin):
            return name_or_type

    if isinstance(name_or_type, Distribution) or name_or_type in Distribution.__members__.values():
        return AllDistributions[name_or_type]
    elif isinstance(name_or_type, str) and name_or_type.upper() in [d.name for d in Distribution]:
        return AllDistributions[Distribution.__members__[name_or_type.upper()]]
    elif isinstance(name_or_type, str) and name_or_type.title() in [d.value for d in Distribution]:
        return AllDistributions[Distribution(name_or_type.title())]
    else:
        return



from enum import Enum, EnumMeta
from typing import Any


class CaseInsensitiveEnumMeta(EnumMeta):
    """
    Enum metaclass to allow for interoperability with case-insensitive strings.
    """

    def __getitem__(cls, name: str) -> Any:
        return super(CaseInsensitiveEnumMeta, cls).__getitem__(name.upper())

    def __getattr__(cls, name: str) -> Enum:
        """Returns the enum member matching `name`.

        We use __getattr__ instead of descriptors or inserting into the enum
        class' __dict__ in order to support `name` and `value` being both
        properties for enum members (which live in the class' __dict__) and
        enum members themselves.

        :param str name: The name of the enum member to retrieve.
        :rtype: ~CaseInsensitiveEnumMeta
        :return: The enum member matching `name`.
        :raises AttributeError: If `name` is not a valid enum member.
        """
        try:
            return cls._member_map_[name.upper()]
        except KeyError as err:
            raise AttributeError(name) from err



from __future__ import annotations

import sys
import warnings
from abc import ABCMeta
from typing import Optional, Sequence

import numpy as np
import pandas as pd
from scipy.special import logsumexp
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils import as_float_array
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import _ensure_no_complex_data, check_is_fitted

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from wnb.stats import Distribution, NonNumericDistributions
from wnb.stats._utils import get_dist_class, is_dist_supported
from wnb.stats.base import DistMixin
from wnb.stats.typing import DistributionLike

from ._utils import (
    SKLEARN_V1_6_OR_LATER,
    _check_feature_names,
    _check_n_features,
    validate_data,
)
from .typing import ArrayLike, Float, MatrixLike

__all__ = ["GeneralNB"]


class GeneralNB(ClassifierMixin, BaseEstimator, metaclass=ABCMeta):
    """A General Naive Bayes classifier that supports distinct likelihood distributions for individual features,
    enabling more tailored modeling beyond the standard single-distribution approaches such as GaussianNB and BernoulliNB.

    Parameters
    ----------
    priors : array-like of shape (n_classes,), default=None
        Prior probabilities of the classes. If specified, the priors are not
        adjusted according to the data.

    distributions : sequence of distribution-like of length n_features, default=None
        Probability distributions to be used for features' likelihoods. If not specified,
        all likelihoods will be considered Gaussian.

    alpha : float, default=1e-10
        Additive (Laplace/Lidstone) smoothing parameter. Set alpha=0 for no smoothing.

    var_smoothing : float, default=1e-9
        Portion of the largest variance of all features that is added to
        variances of normal distributions for calculation stability.

    Attributes
    ----------
    class_count_ : ndarray of shape (n_classes,)
        Number of training samples observed in each class.

    class_prior_ : ndarray of shape (n_classes,)
        Probability of each class.

    classes_ : ndarray of shape (n_classes,)
        Class labels known to the classifier.

    n_classes_ : int
        Number of classes seen during :term:`fit`.

    epsilon_ : float
        Absolute additive value to variances of normal distributions.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    distributions_ : list of length `n_features_in_`
        List of likelihood distributions used to fit to features.

    likelihood_params_ : dict
        A mapping from class labels to their fitted likelihood distributions.
    """

    def __init__(
        self,
        *,
        priors: Optional[ArrayLike] = None,
        distributions: Optional[Sequence[DistributionLike]] = None,
        alpha: Float = 1e-10,
        var_smoothing: Float = 1e-9,
    ) -> None:
        self.priors = priors
        self.distributions = distributions
        self.alpha = alpha
        self.var_smoothing = var_smoothing

    if SKLEARN_V1_6_OR_LATER:

        def __sklearn_tags__(self):
            tags = super().__sklearn_tags__()
            tags.target_tags.required = True
            return tags

    def _more_tags(self) -> dict[str, bool]:
        return {"requires_y": True}

    def _get_distributions(self) -> Sequence[DistributionLike]:
        try:
            if self.distributions_ is not None:
                return self.distributions_
        except Exception:
            return self.distributions or []

    def _check_inputs(self, X, y) -> None:
        # Check if the targets are suitable for classification
        check_classification_targets(y)

        # Check if only one class is present in label vector
        if self.n_classes_ == 1:
            raise ValueError("Classifier can't train when only one class is present")

        X = validate_data(
            self,
            X,
            accept_sparse=False,
            accept_large_sparse=False,
            dtype=(
                None if any(d in self._get_distributions() for d in NonNumericDistributions) else "numeric"
            ),
            force_all_finite=True,
            ensure_2d=True,
            ensure_min_samples=1,
            ensure_min_features=1,
        )

        # Check that the number of samples and labels are compatible
        if X.shape[0] != y.shape[0]:
            raise ValueError("X.shape[0]=%d and y.shape[0]=%d are incompatible." % (X.shape[0], y.shape[0]))

    def _prepare_X_y(self, X=None, y=None, from_fit: bool = False):
        if from_fit and y is None:
            raise ValueError("requires y to be passed, but the target y is None.")

        if X is not None:
            # Convert to NumPy array if X is Pandas DataFrame
            if isinstance(X, pd.DataFrame):
                X = X.values
            _ensure_no_complex_data(X)
            X = (
                X
                if any(d in self._get_distributions() for d in NonNumericDistributions)
                else as_float_array(X)
            )

        if y is not None:
            # Convert to a NumPy array
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
            else:
                y = np.array(y)

            # Warning in case of y being 2d
            if y.ndim > 1:
                warnings.warn(
                    "A column-vector y was passed when a 1d array was expected.",
                    DataConversionWarning,
                )

            y = y.flatten()

        output = tuple(item for item in [X, y] if item is not None)
        return output[0] if len(output) == 1 else output

    def _prepare_parameters(self) -> None:
        self.class_prior_: np.ndarray

        # Set priors if not specified
        if self.priors is None:
            self.class_prior_ = (
                self.class_count_ / self.class_count_.sum()
            )  # Calculate empirical prior probabilities

        else:
            # Check that the provided priors match the number of classes
            if len(self.priors) != self.n_classes_:
                raise ValueError("Number of priors must match the number of classes.")
            # Check that the sum of priors is 1
            if not np.isclose(self.priors.sum(), 1.0):
                raise ValueError("The sum of the priors should be 1.")
            # Check that the priors are non-negative
            if (self.priors < 0).any():
                raise ValueError("Priors must be non-negative.")

            self.class_prior_ = self.priors

        # Convert to NumPy array if input priors is in a list/tuple/set
        if isinstance(self.class_prior_, (list, tuple, set)):
            self.class_prior_ = np.array(list(self.class_prior_))

        # Set distributions if not specified
        if self.distributions is None:
            self.distributions_: list[DistributionLike] = [Distribution.NORMAL] * self.n_features_in_
        else:
            # Check if the number of distributions matches the number of features
            if len(self.distributions) != self.n_features_in_:
                raise ValueError(
                    "Number of specified distributions must match the number of features."
                    f"({len(self.distributions)} != {self.n_features_in_})"
                )

            # Check that all specified distributions are supported
            for i, dist in enumerate(self.distributions):
                if not is_dist_supported(dist):
                    raise ValueError(f"Distribution '{dist}' at index {i} is not supported.")

            self.distributions_: list[DistributionLike] = list(self.distributions)

    def fit(self, X: MatrixLike, y: ArrayLike) -> Self:
        """Fits general Naive Bayes classifier according to X, y.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vectors, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.n_features_in_: int
        self.feature_names_in_: np.ndarray
        _check_n_features(self, X=X, reset=True)
        _check_feature_names(self, X=X, reset=True)

        X, y = self._prepare_X_y(X, y, from_fit=True)

        self.classes_: np.ndarray
        self.class_count_: np.ndarray
        self.classes_, y_, self.class_count_ = np.unique(
            y, return_counts=True, return_inverse=True
        )  # Unique class labels, their indices, and class counts
        self.n_classes_: int = len(self.classes_)  # Number of classes

        self._check_inputs(X, y)
        y = y_
        self._prepare_parameters()

        self.epsilon_ = 0.0
        if np.all(np.isreal(X)):
            self.epsilon_ = self.var_smoothing * np.var(X, axis=0).max()

        self.likelihood_params_: dict[int, list[DistMixin]] = {
            c: [
                get_dist_class(self.distributions_[i]).from_data(
                    X[y == c, i], alpha=self.alpha, epsilon=self.epsilon_
                )
                for i in range(self.n_features_in_)
            ]
            for c in range(self.n_classes_)
        }

        return self

    def predict(self, X: MatrixLike) -> np.ndarray:
        """Performs classification on an array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : ndarray of shape (n_samples,)
            Predicted target values for X.
        """
        p_hat = self.predict_log_proba(X)
        return self.classes_[np.argmax(p_hat, axis=1)]

    def predict_log_proba(self, X: MatrixLike) -> np.ndarray:
        """Returns log-probability estimates for the array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : array-like of shape (n_samples, n_classes)
            Returns the log-probability of the samples for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute :term:`classes_`.
        """
        # Check is fit had been called
        check_is_fitted(self)

        # Input validation
        X = validate_data(
            self,
            X,
            accept_large_sparse=False,
            force_all_finite=True,
            dtype=(
                None if any(d in self._get_distributions() for d in NonNumericDistributions) else "numeric"
            ),
            reset=False,
        )

        # Check if the number of input features matches the data seen during fit
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Expected input with %d features, got %d instead." % (self.n_features_in_, X.shape[1])
            )

        n_samples = X.shape[0]
        X = self._prepare_X_y(X=X)

        log_joint = np.zeros((n_samples, self.n_classes_))
        for c in range(self.n_classes_):
            log_joint[:, c] = np.log(self.class_prior_[c]) + np.sum(
                [np.log(likelihood(X[:, i])) for i, likelihood in enumerate(self.likelihood_params_[c])],
                axis=0,
            )

        log_proba = log_joint - np.transpose(
            np.repeat(logsumexp(log_joint, axis=1).reshape(1, -1), self.n_classes_, axis=0)
        )
        return log_proba

    def predict_proba(self, X: MatrixLike) -> np.ndarray:
        """Returns probability estimates for the array of test vectors X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        C : array-like of shape (n_samples, n_classes)
            Returns the probability of the samples for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute :term:`classes_`.
        """
        return np.exp(self.predict_log_proba(X))



from typing import Any

import sklearn
from packaging import version
from sklearn.utils import check_array

__all__ = [
    "SKLEARN_V1_6_OR_LATER",
    "validate_data",
    "_check_n_features",
    "_check_feature_names",
]


SKLEARN_V1_6_OR_LATER = version.parse(sklearn.__version__) >= version.parse("1.6")


if SKLEARN_V1_6_OR_LATER:
    from sklearn.utils.validation import _check_feature_names, _check_n_features
    from sklearn.utils.validation import validate_data as _validate_data

    def validate_data(*args, **kwargs):
        if kwargs.get("force_all_finite"):
            kwargs["ensure_all_finite"] = kwargs.pop("force_all_finite")
        return _validate_data(*args, **kwargs)

else:

    def validate_data(estimator, X, **kwargs: Any):
        kwargs.pop("reset", None)
        return check_array(X, estimator=estimator, **kwargs)

    def _check_n_features(estimator, X, reset):
        return estimator._check_n_features(X, reset=reset)

    def _check_feature_names(estimator, X, *, reset):
        return estimator._check_feature_names(X, reset=reset)



from __future__ import annotations

import warnings

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import clone
from sklearn.model_selection import train_test_split

try:
    from tqdm import tqdm
except ModuleNotFoundError:

    def tqdm(iterable, *args, **kwargs):
        return iterable


warnings.filterwarnings("ignore")

__all__ = ("benchmark",)


def compare_score(X, y, wnb, sklearn, random_state: int, test_size: float) -> tuple[np.float64, np.float64]:
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, stratify=y, random_state=random_state
    )

    clf_wnb = clone(wnb)
    clf_wnb.fit(X_train, y_train)

    clf_sklearn = clone(sklearn)
    clf_sklearn.fit(X_train, y_train)

    return clf_wnb.score(X_test, y_test), clf_sklearn.score(X_test, y_test)


def benchmark(
    X, y, wnb, sklearn, max_iter: int = 50, test_size: float = 0.33
) -> tuple[np.float64, np.float64]:
    results = Parallel(n_jobs=-1, prefer="processes")(
        delayed(compare_score)(*param)
        for param in tqdm([(X, y, wnb, sklearn, i, test_size) for i in range(max_iter)], ncols=80)
    )

    return np.mean([r[0] for r in results]), np.mean([r[1] for r in results])
